import Rappture
import sys
from volume_calculations.isopach import Isopach
from volume_calculations import model_exp, model_pow, model_wei
from volume_calculations import regression_methods
from volume_calculations.exceptions import ExpectedException

import math

import numpy
numpy.seterr(over='raise')

io = Rappture.library(sys.argv[1])

NPTS = 100
X_MAX_SCALE = 1.2

SF = 4

titleBar = "-----------------"

def main():

	modelType = io.get('input.phase(model).choice(modelType).current')
	
	isopachs = loadIsopachsFromFile()
	isopachs = sorted(isopachs, key=lambda isopach : isopach.sqrtAreaKM)

	io.put('output.curve(modelData).about.label','Model',append=0)
	io.put('output.curve(modelData).about.description','Isopach data',append=0)
	io.put('output.curve(modelData).about.type','scatter',append=0)
	io.put('output.curve(modelData).group','Model',append=0)
	io.put('output.curve(modelData).xaxis.label','sqrt(area)',append=0)
	io.put('output.curve(modelData).xaxis.units','km',append=0)
	io.put('output.curve(modelData).yaxis.label','thickness',append=0)
	io.put('output.curve(modelData).yaxis.units','m',append=0)
	
	if modelType == "Exponential":
		displayExponential(isopachs,modelType)

	elif modelType == "Power law":
		displayPowerLaw(isopachs,modelType)

	elif modelType == "Weibull":
		displayWeibull(isopachs,modelType)

	Rappture.result(io)

def loadIsopachsFromFile():

	data = io.get('input.phase(isopachs).string(file).current')
	
	if data == "":
		raise ExpectedException('Please input data using the "Isopachs" tab')

	splitLines = [line.split(",") for line in data.split('\n')]

	isopachs = []
	for index,line in enumerate(splitLines):
		if len(line) != 2:
			raise ExpectedException("Expected two values per line, found " + str(len(line)) + " in line " + str(index+1))
		else:
			try:
				thicknessM = float(line[0])
			except FloatingPointException:
				raise ExpectedException("Could not parse " + line[0] + " as a number in line " + str(index))

			try:
				sqrtAreaKM = float(line[1])
			except FloatingPointException:
				raise ExpectedException("Could not parse " + line[1] + " as a number in line " + str(index))

			if thicknessM == 0 or sqrtAreaKM == 0:
				raise ExpectedException("Isopach values must be non-zero")
				
			isopachs.append(Isopach(sqrtAreaKM,thicknessM))

	return isopachs

def displayExponential(isopachs,modelType):

	segments = int(io.get('input.phase(model).group(expParameters).integer(segments).current'))

	n = len(isopachs)
	if n < segments*2:
		raise ExpectedException("Cannot perform linear regression for " + str(segments) + " segment" +
								(" " if segments == 1 else "s ") +
								"with only " + str(n) + " unique x-value" + (" " if n == 1 else "s "))

 	results = model_exp.exponentialModelAnalysis(isopachs,segments)

	estimatedTotalVolumeStr = roundToSF(results['estimatedTotalVolume'],SF)
	segmentLimits = results['segmentLimits']
	segmentVolumes = results['segmentVolumes']
	segmentCoefficients = results['segmentCoefficients']
	segmentExponents = results['segmentExponents']
	regressionLines = results['regressionLines']
	thicknessFunction = results['thicknessFunction']
	mrse = regression_methods.meanRelativeSquaredError([isopach.sqrtAreaKM for isopach in isopachs],
													   [isopach.thicknessM for isopach in isopachs],
													   thicknessFunction)
	mrseStr = roundToSF(mrse,SF)

	output = (titleBar + '- Results -' + titleBar + '\n\n'
			'  Estimated total volume: ' + estimatedTotalVolumeStr + 'km3\n\n\n'
			'  Mean relative squared error: ' + mrseStr + '\n\n')

	for i in range(segments):
		
		coefficientStr = roundToSF(segmentCoefficients[i],SF)
		exponentStr = roundToSF(segmentExponents[i],SF)
		nExponentStr = roundToSF(-segmentExponents[i],SF)
		segVolumeStr = roundToSF(segmentVolumes[i],SF)
		lowerLimitStr = roundToSF(segmentLimits[i],SF)
		upperLimitStr = roundToSF(segmentLimits[i+1],SF)

		output += ('\n  Segment ' + str(i+1) + '\n\n'
					'	  Volume: ' + segVolumeStr + 'km3\n'
					'	  Range: ' + lowerLimitStr + '-' + upperLimitStr + '\n'
					'	  c: ' + coefficientStr + '\n'
					'	  m: ' + exponentStr + '\n'
					'	  Equation: T(x) = ' + coefficientStr + '*exp(' + nExponentStr + 'x)\n')
	
	output += '\n\n' + titleBar + ' Parameters ' + titleBar + '\n\n' + '  Model type: ' + modelType + '\n' + '  Segments: ' + str(segments) + '\n'

	io.put('output.log(expNumericalResults)',output,append=0)
	io.put('output.log(expNumericalResults).about.label','Numerical',append=0)
	io.put('output.log(expNumericalResults).about.description',"Exponential model results\n\nEstimated total volume - the total volume of the deposit in km3 as predicted by the model, obtained by integrating the thickness function with respect to area between 0 and inf.\n\nMRSE - mean relative squared error, measures how well the model fits the data. A score of zero denotes a perfect fit. Note that accurate comparisons can only be made between models when they use the same isopach data.\n\nSegments\n\nVolume - the volume of the segment in km3\nRange - the range of x-values for which the segment is valid.\nc - the value of the segment's coefficient, c.\nm - the value of segment's exponent, m.\nEquation - the thickness equation for the segment, only valid for x values in the segment's range.",append=0)

	for isopach in isopachs:
		io.put('output.curve(modelData).component.xy',
				str(isopach.sqrtAreaKM)+' '+str(isopach.thicknessM)+'\n',
				append=1)
	
	for seg in range(segments):
		
		io.put('output.curve(modelCurve'+str(seg)+').about.group','Model',append=0)
		io.put('output.curve(modelCurve'+str(seg)+').about.label','Segment ' + str(seg+1),append=0)
		
		c = segmentCoefficients[seg]
		m = segmentExponents[seg]

		xmin = segmentLimits[seg]
		xmax = segmentLimits[seg+1] if seg != segments-1 else isopachs[-1].sqrtAreaKM*X_MAX_SCALE
	
		xs = [(xmin+1)*((1.0*(xmax+1)/(xmin+1))**(1.0*i/(NPTS)))-1 for i in range(NPTS+1)]

		for x in xs:
			io.put('output.curve(modelCurve'+str(seg)+').component.xy',
					str(x)+' '+str(c*math.exp(-m*x))+'\n',
					append=1)
		#ys = [c*math.exp(-x*m) for x in xs]

	io.put('output.curve(regressionData).about.group','Regression',append=0)
	io.put('output.curve(regressionData).about.label','Regression data',append=0)
	io.put('output.curve(regressionData).about.type','scatter',append=0)
	io.put('output.curve(regressionData).xaxis.label','sqrt(area)',append=0)
	io.put('output.curve(regressionData).yaxis.label','log(thickness)',append=0)

	regressionDataXs = [isopach.sqrtAreaKM for isopach in isopachs]
	regressionDataYs = [math.log(isopach.thicknessM) for isopach in isopachs]
	
	for i in range(segments):
		x1 = 0 if i == 0 else segmentLimits[i]
		y1 = math.log(thicknessFunction(x1))

		if i == segments-1:
			x2 = isopachs[-1].sqrtAreaKM*X_MAX_SCALE
		else:
			x2 = segmentLimits[i+1]
		y2 = math.log(thicknessFunction(x2))

		io.put('output.curve(regressionLine'+str(i+1)+').about.group','Regression',append=0)
		io.put('output.curve(regressionLine'+str(i+1)+').about.label','Regression line',append=0)
		io.put('output.curve(regressionLine'+str(i+1)+').about.color','green',append=0)

		io.put('output.curve(regressionLine'+str(i+1)+').component.xy',
				roundToSF(x1,SF) + ' '  + roundToSF(y1,SF) + '\n' + roundToSF(x2,SF) + ' '  + roundToSF(y2,SF) + '\n',
				append=1)

	for x, y in zip(regressionDataXs,regressionDataYs):
		io.put('output.curve(regressionData).component.xy',
				roundToSF(x,SF) + ' '  + roundToSF(y,SF) + '\n',
				append=1)
	
	

def displayPowerLaw(isopachs,modelType):

	proximalLimit = float(io.get('input.phase(model).group(powParameters).number(proximalLimit).current'))
	distalLimit = float(io.get('input.phase(model).group(powParameters).number(distalLimit).current'))
	
	if proximalLimit >= distalLimit:
		raise ExpectedException('The proximal limit must be less than or equal to distal limit')

	results = model_pow.powerLawModelAnalysis(isopachs,proximalLimit,distalLimit)
	
	estimatedTotalVolume = results['estimatedTotalVolume']
	coefficient = results['coefficient']
	exponent = results['exponent']
	thicknessFunction = results['thicknessFunction']
	regressionLine = results['regressionLine']
	mrse = regression_methods.meanRelativeSquaredError([isopach.sqrtAreaKM for isopach in isopachs],
													   [isopach.thicknessM for isopach in isopachs],
													   thicknessFunction)
	proximalLimitEstimate = model_pow.calculateProximalLimitEstimate(isopachs,coefficient,exponent)

	estimatedTotalVolumeStr = roundToSF(estimatedTotalVolume,SF)
	coefficientStr = roundToSF(coefficient,SF)
	exponentStr = roundToSF(exponent,SF)
	nExponentStr = roundToSF(-exponent,SF)
	mrseStr = roundToSF(mrse,SF)
	proximalLimitEstimateStr = roundToSF(proximalLimitEstimate,SF)	

	equationStr = 'T(x) = '
	if coefficient == 0.0:
		equationStr += '0.0'
	else:
		if coefficient != 1.0:
			equationStr += coefficientStr
		
		if exponent != 0.0:
			equationStr += '*x'

			if exponent != -1.0:
				equationStr += '^' + nExponentStr

	output = (titleBar + '- Results -' + titleBar + '\n\n'
			'  Estimated total volume: ' + estimatedTotalVolumeStr + 'km3\n\n\n'
			'  Mean relative squared error: ' + mrseStr + '\n\n'
			'  Coefficient: ' + coefficientStr + '\n'
			'  Exponent: ' + exponentStr + '\n'			
			'  Equation: ' + equationStr + '\n\n'
			'  Estimated proximal limit: ' + proximalLimitEstimateStr + '\n\n\n'
			'' + titleBar + ' Parameters ' + titleBar + '\n\n' + '  Model type: ' + modelType + '\n\n'
			'  Proximal limit: ' + str(proximalLimit) + '\n'
			'  Distal limit: ' + str(distalLimit) + '\n')

	io.put('output.log(powNumericalResults)',output,append=0)
	io.put('output.log(powNumericalResults).about.label','Numerical',append=0)
	io.put('output.log(powNumericalResults).about.description','Power law model results\n\nEstimated total volume - the total volume of the deposit in km3 as predicted by the model, obtained by integrating the thickness function with respect to area between 0 and inf.\n\nMRSE - mean relative squared error, measures how well the model fits the data. A score of zero denotes a perfect fit. Note that accurate comparisons can only be made between models when they use the same isopach data.\n\nc - the value of the coefficient, c, to ' + str(SF) + ' s.f.\nm - the value of the exponent, m, to ' + str(SF) + ' s.f.\nEquation - the thickness equation for the model.\n\nEstimated proximal limit - an estimate of the proximal limit derived from the isopach data using the method described in Bonadonna and Houghton (2005).',append=0)

	io.put('output.curve(modelCurve).about.label','Model',append=0)
	io.put('output.curve(modelCurve).about.description','Isopach data',append=0)
	io.put('output.curve(modelCurve).about.group','Model',append=0)
	io.put('output.curve(modelCurve).xaxis.label','sqrt(area)',append=0)
	io.put('output.curve(modelCurve).xaxis.units','km',append=0)
	io.put('output.curve(modelCurve).yaxis.label','thickness',append=0)
	io.put('output.curve(modelCurve).yaxis.units','m',append=0)

	xmin = isopachs[0].sqrtAreaKM/X_MAX_SCALE
	xmax = isopachs[-1].sqrtAreaKM*X_MAX_SCALE

	modelDataXs = [isopach.sqrtAreaKM for isopach in isopachs]
	modelDataYs = [isopach.thicknessM for isopach in isopachs]

	modelCurveXs = [xmin*((1.0*xmax/xmin)**(1.0*i/(NPTS+1))) for i in range(NPTS+1)]
	modelCurveYs = [thicknessFunction(x) for x in modelCurveXs]

	for x, y in zip(modelCurveXs,modelCurveYs):
		io.put('output.curve(modelCurve).component.xy', 
				str(roundToSF(x,SF)) + ' ' + str(roundToSF(y,SF)) + '\n', 
				append=1)

	for x,y in zip(modelDataXs,modelDataYs):
		io.put('output.curve(modelData).component.xy', 
				str(x) + ' ' + str(y) + '\n',
				append=1)


	io.put('output.curve(regressionData).about.group','Regression',append=0)
	io.put('output.curve(regressionData).about.label','Regression line',append=0)
	io.put('output.curve(regressionData).about.type','scatter',append=0)
	io.put('output.curve(regressionData).xaxis.label','log(sqrt(area))',append=0)
	io.put('output.curve(regressionData).yaxis.label','log(thickness)',append=0)

	io.put('output.curve(regressionLine).about.group','Regression',append=0)
	io.put('output.curve(regressionLine).about.label','Data',append=0)
	io.put('output.curve(regressionLine).xaxis.label','log(sqrt(area))',append=0)
	io.put('output.curve(regressionLine).yaxis.label','log(thickness)',append=0)

	regressionDataXs = [math.log(isopach.sqrtAreaKM) for isopach in isopachs]
	regressionDataYs = [math.log(isopach.thicknessM) for isopach in isopachs]
	
	regressionLineXs = [regressionDataXs[0]/1.2,regressionDataXs[-1]*1.2]
	regressionLineYs = [regressionLine.m*x+regressionLine.c for x in regressionLineXs]

	for x, y in zip(regressionDataXs,regressionDataYs):
		io.put('output.curve(regressionData).component.xy',
				roundToSF(x,SF) + ' '  + roundToSF(y,SF) + '\n',
				append=1)
	
	for x, y in zip(regressionLineXs,regressionLineYs):
		io.put('output.curve(regressionLine).component.xy',
				roundToSF(x,SF) + ' '  + roundToSF(y,SF) + '\n',
				append=1)

def displayWeibull(isopachs,modelType):

	runs = int(io.get('input.phase(model).group(weiParameters).integer(runs).current'))
	iterationsPerRun = int(io.get('input.phase(model).group(weiParameters).integer(iterationsPerRun).current'))
	lambdaLower = float(io.get('input.phase(model).group(weiParameters).number(lambdaLower).current'))
	lambdaUpper = float(io.get('input.phase(model).group(weiParameters).number(lambdaUpper).current'))
	kLower = float(io.get('input.phase(model).group(weiParameters).number(kLower).current'))
	kUpper = float(io.get('input.phase(model).group(weiParameters).number(kUpper).current'))
	
	if lambdaLower >= lambdaUpper:
		raise ExpectedException('The lower bound for lambda must be less than or equal to the upper bound for lambda')
	if kLower >= kUpper:
		raise ExpectedException('The lower bound for k must be less than or equal ot the upper bound for k')


	results = model_wei.weibullModelAnalysis(isopachs,
											runs,
											iterationsPerRun,
											((lambdaLower,lambdaUpper),(kLower,kUpper)))
	
	estimatedTotalVolume = results['estimatedTotalVolume']
	lamb = results['lambda']
	k = results['k']
	theta = results['theta']
	thicknessFunction = results['thicknessFunction']
	mrse = regression_methods.meanRelativeSquaredError([isopach.sqrtAreaKM for isopach in isopachs],
													   [isopach.thicknessM for isopach in isopachs],
													   thicknessFunction)

	estimatedTotalVolumeStr = roundToSF(estimatedTotalVolume,SF)
	lambdaStr = roundToSF(lamb,SF)
	inverseLambdaStr = roundToSF(1/lamb,SF)
	kStr = roundToSF(k,SF)
	kminus2Str = roundToSF(k-2,SF)
	thetaStr = roundToSF(theta,SF)
	mrseStr = roundToSF(mrse,SF)	

	equationStr = 'T(x) = ' + thetaStr + '((' + inverseLambdaStr + 'x)^' +kminus2Str + ')exp(-(' + inverseLambdaStr + 'x)^' + kStr + ')'
 
	output = (titleBar + '- Results -' + titleBar + '\n\n'
			'  Estimated total volume: ' + estimatedTotalVolumeStr + 'km3\n\n\n'
			'  Mean relative squared error: ' + mrseStr + '\n\n'
			'  lambda: ' + lambdaStr + '\n'
			'  k: ' + kStr + '\n'	
			'  theta: ' + thetaStr + '\n\n'
			'  Equation: ' + equationStr + '\n\n\n'
			'' + titleBar + ' Parameters ' + titleBar + '\n\n' + '  Model type: ' + modelType + '\n\n'
			'  Number of runs: ' + str(runs) + '\n'
			'  Iterations per run: ' + str(iterationsPerRun) + '\n\n'
			'  lambda lower bound: ' + roundToSF(lambdaLower,SF) + '\n'
			'  lambda upper bound: ' + roundToSF(lambdaUpper,SF) + '\n'
			'  k lower bound: ' + roundToSF(kLower,SF) + '\n'
			'  k upper bound: ' + roundToSF(kUpper,SF) + '\n')

	io.put('output.log(weiNumericalResults)',output,append=1)
	io.put('output.log(weiNumericalResults).about.label','Numerical',append=0)
	io.put('output.log(weiNumericalResults).about.description','Weibull model results\n\nNOTE: the parameters output for the Weibull model are generated using a non-deterministic algorithm and hence different models may be generated each time. The models obtained can be compared using the MRSE value.\n\nEstimated total volume - the total volume of the deposit in km3 as predicted by the model, obtained by integrating the thickness function with respect to area between 0 and inf.\n\nMRSE - mean relative squared error, measures how well the model fits the data. A score of zero denotes a perfect fit. Note that accurate comparisons can only be made between models when they use the same isopach data.\n\nlambda - the value of lambda to ' + str(SF) + ' s.f.\nk - the value of k to ' + str(SF) + ' s.f.\ntheta - the value of theta to ' + str(SF) + ' s.f.\nEquation - the thickness equation for the model.',append=0)

	io.put('output.curve(modelCurve).about.label','Model',append=0)
	io.put('output.curve(modelCurve).about.description','Isopach data',append=0)
	io.put('output.curve(modelCurve).about.group','Model',append=0)
	io.put('output.curve(modelCurve).xaxis.label','sqrt(area)',append=0)
	io.put('output.curve(modelCurve).xaxis.units','km',append=0)
	io.put('output.curve(modelCurve).yaxis.label','thickness',append=0)
	io.put('output.curve(modelCurve).yaxis.units','m',append=0)

	xmin = isopachs[0].sqrtAreaKM/X_MAX_SCALE
	xmax = isopachs[-1].sqrtAreaKM*X_MAX_SCALE

	modelDataXs = [isopach.sqrtAreaKM for isopach in isopachs]
	modelDataYs = [isopach.thicknessM for isopach in isopachs]

	modelCurveXs = [xmin*((1.0*xmax/xmin)**(1.0*i/(NPTS+1))) for i in range(NPTS+1)]
	modelCurveYs = [thicknessFunction(x) for x in modelCurveXs]

	for x, y in zip(modelCurveXs,modelCurveYs):
		io.put('output.curve(modelCurve).component.xy', 
				str(roundToSF(x,SF)) + ' ' + str(roundToSF(y,SF)) + '\n', 
				append=1)

	for x,y in zip(modelDataXs,modelDataYs):
		io.put('output.curve(modelData).component.xy', 
				str(x) + ' ' + str(y) + '\n',
				append=1)

	graphPSpace = io.get('input.phase(model).group(pSpace).boolean(graphPSpace).current')

	if graphPSpace == 'yes':
		res = int(io.get('input.phase(model).group(pSpace).group(weibullGraphParameters).integer(weiResolution).current'))

		if lambdaLower == 0:
			lambdaLower = lambdaUpper/res
			res -= 1

		def errorFunction(lamb,k):

			theta = model_wei.calculateTheta(modelDataXs,modelDataYs,lamb,k)
			def thicknessFunction(x):
				if lamb == 0:
					return 0

				try:
					return math.exp(math.log(theta)+(k-2)*math.log(x/lamb)-(x/lamb)**k)
				except FloatingPointError, ValueError:
					return 0

			return math.log(regression_methods.meanRelativeSquaredError(modelDataXs, modelDataYs, thicknessFunction))
		
		io.put('output.mesh(weiMesh).about.label','Weibull mesh',append=0)		
		io.put('output.mesh(weiMesh).dim',2,append=0)
		io.put('output.mesh(weiMesh).hide','yes',append=0)		
		io.put('output.mesh(weiMesh).grid.xaxis.min',lambdaLower,append=0)
		io.put('output.mesh(weiMesh).grid.xaxis.max',lambdaUpper,append=0)
		io.put('output.mesh(weiMesh).grid.xaxis.numpoints',res,append=0)
		io.put('output.mesh(weiMesh).grid.yaxis.min',kLower,append=0)
		io.put('output.mesh(weiMesh).grid.yaxis.max',kUpper,append=0)
		io.put('output.mesh(weiMesh).grid.yaxis.numpoints',res,append=0)
		
		io.put('output.log(numericalResults)',str(lambdaLower)+" "+str(lambdaUpper)+" "+str(kLower)+" "+str(kUpper), append=1)
		io.put('output.field(weiField).about.label','Weibull error surface',append=0)
		io.put('output.field(weiField).component.mesh','output.mesh(weiMesh)',append=0)
		io.put('output.field(weiField).about.xaxis.label','lambda',append=0)
		io.put('output.field(weiField).about.yaxis.label','k',append=0)
		io.put('output.field(weiField).about.zaxis.label','MRSE',append=0)

		lambdaWidth = lambdaUpper-lambdaLower
		kWidth = kUpper-kLower
		for i in range(res):
			for j in range(res):
				lambdax = lambdaLower+j*lambdaWidth/res
				ky = kLower+i*kWidth/res
				errorz = errorFunction(lambdax,ky)

				io.put('output.field(weiField).component.values',str(errorz) + '\n',append=1)

def roundToSF(number,sf):
	if number == 0:
		return '0.'+'0'*sf
	elif number == float('inf'):
		return  "inf"
	else:
		places = int(math.floor(math.log10(abs(number))))
		value = round(number, -places + (sf - 1))

		if places >= sf - 1:
			string = str(int(value))
		else:
			string = str(value)
			if len(string) <= sf:
				string += "0"*(sf-len(string)+1)
			
		return string

try:
	main()
except ExpectedException as ee:
	raise ExpectedException('\n\nExpected error\n\n' + ee.message)
except Exception as e:
	raise Exception('\n\nUnexpected error, please consider reporting it on VHub\n\n' + e.message)
